unit NkVImage;

{$IMPORTEDDATA OFF}
{$MINENUMSIZE 4}

interface

uses Windows, ImgFilter;

type
  TMaxBitmapInfo = packed record
    bmiHeader: TBitmapInfoHeader;
    bmiColors: array [0..255+3] of TRGBQuad;
  end;

  TPlugin = record
    Path: WideString;
    UseExtInfo: Boolean;
  end;
  TPlugins = array of TPlugin;

  TNkvImageFormat = (nifBMP, nifJPEG, nifPNG);

  TNkvImage = class(TObject)
  private
    FBitmapInfo: TMaxBitmapInfo;
    FBits: PChar;
    procedure Scale(var DestBmi: TMaxBitmapInfo; DestBits: Pointer;
      DestWidth, DestHeight, ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer;
      Filter: Boolean);
  public
    Loaded: Boolean;
    Width: Integer;
    Height: Integer;
    BitsSize: Cardinal;
    Bpp: Cardinal;
    OrgSize: Cardinal;
    FormatName: string;
    ICCProfileSupported: Boolean;
    ICCProfileName: string;
    Comment: string;
    FilterName: string;
    FilterFileName: WideString;
    procedure ReadMem(Buf: PChar; BufSize: Cardinal; const Filter: TImgFilter);
    procedure Draw(hDC: HDC; Left, Top, AWidth, AHeight: Integer;
      DestLeft, DestTop, DestWidth, DestHeight: Integer;
      Grid: Boolean; ScaleFilter: Boolean; GridSize: Integer);
    procedure Write(const FileName: WideString; Format: TNkvImageFormat;
      AWidth, AHeight: Integer; ScaleFilter: Boolean);
    procedure GetPixel(X, Y: Integer; var R, G, B: Integer);
  end;

function LookupFilter(const AFileName: string; Buf: PChar; BufSize: Integer;
    const Plugins: TPlugins): TObject;

implementation

uses NLib, NSys, ImgScaler, SpiFilter, ifpnm, ifdib, ifgdiplus, ifwic;

function LookupFilter(const AFileName: string; Buf: PChar; BufSize: Integer;
    const Plugins: TPlugins): TObject;
const
  Filters: array [0..1] of TImgFilterClass = (TDibFilter, TPnmFilter);
var
  i: Integer;
begin
  for i := 0 to High(Filters) do
  begin
    Result := Filters[i].Create;
    if (Result as TImgFilter).Test(Buf, BufSize) then
      Exit;
    Result.Free;
  end;

  for i := 0 to High(Plugins) do
  begin
    Result := TSpiFilter.Create;
    with Result as TSpiFilter do
    begin
      LoadPlugin(Plugins[i].Path, Plugins[i].UseExtInfo);
      if TestSpi(AFileName, Buf, BufSize) then
      begin
        if IsArchive then
        begin
          Result.Free;
          Result := TArchive.Create(Plugins[i].Path);
        end;
        Exit;
      end;
      Result.Free;
    end;
  end;

  if GdiplusEnabled then
  begin
    Result := TGdiplusFilter.Create;
    if (Result as TImgFilter).Test(Buf, BufSize) then
      Exit;
    Result.Free;
  end;

  Result := TWicFilter.Create;
  if (Result as TImgFilter).Test(Buf, BufSize) then
    Exit;
  Result.Free;

  Result := nil;
end;

function GetImageFormat(Buf: PChar; BufSize: Cardinal): string;
  function match(p: PChar; ofs, len: Cardinal): Boolean;
  begin
    Result := ((ofs + len) <= BufSize) and (MemCmp(@Buf[ofs], p, len) = 0);
  end;
begin
  if match('GIF87a', 0, 6) then
    Result := 'GIF87a'
  else if match('GIF89a', 0, 6) then
    Result := 'GIF89a'
  else if match(#$FF#$D8, 0, 2) then begin
    if match('JFIF'#0, 6, 5) then
      Result:= 'JPEG (JFIF)'
    else if match('Exif'#0, 6, 5) then
      Result := 'JPEG (Exif)'
    else
      Result := 'JPEG';
  end else if match(#$89'PNG'#$0D#$0A#$1A#$0A, 0, 8) then begin
    Result := 'PNG';
    case Buf[28] of
      #0: Result := Result + ' non-interlaced';
      #1: Result := Result + ' interlaced';
    end;
    if Ord(Buf[25]) and 4 <> 0 then
      Result := Result + ' with alpha';
  end else if match('MM'#$00#$2a, 0, 4) then
    Result := 'TIFF big-endian'
  else if match('II'#$2a#$00, 0, 4) then
    Result := 'TIFF little-endian'
  else if match('RIFF', 0, 4) and match('WEBP', 8, 4) then begin
    if match('VP8 ', 12, 4) then
      Result := 'WebP VP8'
    else if match('VP8L', 12, 4) then
      Result := 'WebP VP8L'
    else if match('VP8X', 12, 4) then begin
      Result := 'WebP VP8X     ';
      Move(Buf[30], Result[11], 4)
    end else
      Result := '';
  end else
    Result := '';
end;

procedure TNkVImage.ReadMem(Buf: PChar; BufSize: Cardinal; const Filter: TImgFilter);
var
  i: Integer;
  bmi: TMaxBitmapInfo;
  tmpbits: PChar;
  imginfo: TImgInfo;
  size: Cardinal;
begin
  tmpbits := nil;
  size := 0;
  try
    try
      Filter.Decode(Buf, BufSize, tmpbits, size, imginfo);
    except 
      on E: EImgFilterError do
        raise EImgFilterError.Create(Filter.Name+':'#13#10+E.Msg);
    end;
  except
    FreeBuf(tmpbits);
    raise;
  end;

  FillChar(bmi.bmiHeader, SizeOf(bmi.bmiHeader), 0);
  with bmi.bmiHeader do
  begin
    biSize := SizeOf(bmi.bmiHeader);
    biWidth := imginfo.Width;
    biHeight := imginfo.Height;
    biPlanes := 1;
    biBitCount := imginfo.Components * 8;
    biCompression := BI_RGB;
    biSizeImage := size;
  end;

  if imginfo.Components = 1 then // gray scale
    for i := 0 to 255 do
    begin
      bmi.bmiColors[i].rgbBlue := i;
      bmi.bmiColors[i].rgbGreen := i;
      bmi.bmiColors[i].rgbRed := i;
      bmi.bmiColors[i].rgbReserved := 0;
    end;

  FreeBuf(FBits);
  FBitmapInfo := bmi;
  FBits := tmpbits;
  BitsSize := size;
  Width := imginfo.Width;
  Height := imginfo.Height;
  Bpp := imginfo.bpp;
  if imginfo.FormatName = '' then
    FormatName := GetImageFormat(Buf, BufSize)
  else
    FormatName := imginfo.FormatName;

  ICCProfileSupported := (Filter is TGdiplusFilter) or (Filter is TWicFilter);
  ICCProfileName := imginfo.ICCProfileName;

  Comment := imginfo.Comment;
  FilterName := Filter.Name;
  if Filter is TSpiFilter then
  begin
    SetString(FilterFileName, nil, MAX_PATH);
    SetLength(FilterFileName, GetModuleFileNameW((Filter as TSpiFilter).Handle,
      PWideChar(FilterFileName), Length(FilterFileName)));
  end else
    FilterFileName := '';
  OrgSize := BufSize;
  Loaded := True;
end;

procedure TNkVImage.Scale(var DestBmi: TMaxBitmapInfo; DestBits: Pointer;
  DestWidth, DestHeight, ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer;
  Filter: Boolean);
begin
  ViewTop := DestHeight - ViewTop - ViewHeight; // for bottom-up DIB
  with FBitmapInfo.bmiHeader do
    if Filter then
    begin
      if (DestWidth > biWidth) or (biWidth div DestWidth <= 2) then
        Bilinear(FBits, DestBits, biWidth, biHeight, biBitCount,
          DestWidth, DestHeight, ViewLeft, ViewTop, ViewWidth, ViewHeight)
      else
        Average(FBits, DestBits, biWidth, biHeight, biBitCount,
          DestWidth, DestHeight, ViewLeft, ViewTop, ViewWidth, ViewHeight);
    end else
      Nearest(FBits, DestBits, biWidth, biHeight, biBitCount,
        DestWidth, DestHeight, ViewLeft, ViewTop, ViewWidth, ViewHeight);

  DestBmi := FBitmapInfo;
  DestBmi.bmiHeader.biWidth := ViewWidth;
  DestBmi.bmiHeader.biHeight := ViewHeight;
end;

procedure TNkVImage.Draw(hDC: HDC; Left, Top, AWidth, AHeight: Integer;
  DestLeft, DestTop, DestWidth, DestHeight: Integer; Grid: Boolean;
  ScaleFilter: Boolean; GridSize: Integer);
var
  bmi: TMaxBitmapInfo;
  tmpbits: Pointer;
  x, y, pitch, a, b, c: Integer;
  z: Double;
  tmp: PChar;
begin
  if (DestWidth = Width) and (DestHeight = Height) and not Grid then
    SetDIBitsToDevice(hDC, Left, Top, DestWidth, DestHeight, DestLeft, 0,
      DestTop, DestHeight, FBits, PBitmapInfo(@FBitmapInfo)^, DIB_RGB_COLORS)
  else if (DestHeight <> 0) and (DestWidth <> 0) then
  begin
    pitch := CalcPitch(AWidth, FBitmapInfo.bmiHeader.biBitCount);

    // If Delphi's memory manager returned splitted memory block,
    // SetDIBitsToDevice() may fail. So use GlobalAlloc().
    tmpbits := Pointer(GlobalAlloc(GMEM_FIXED, pitch * AHeight));
    Scale(bmi, tmpbits, DestWidth, DestHeight, DestLeft, DestTop,
      AWidth, AHeight, ScaleFilter);

    if Grid then
    begin
      c := 0;
      z := Width / DestWidth;
      b := Trunc(z * (DestLeft-1+0.5));
      for x := 0 to AWidth - 1 do
      begin
        a := Trunc(z * (DestLeft+x+0.5));
        if b <> a then
        begin
          b := a;
          if z > 0.25 then
          begin
            if b mod (GridSize*2) = 0 then
              c := 0
            else if b mod GridSize = 0 then
              c := 128
            else
              continue;
          end else
            if b mod GridSize = 0 then
              c := 0
            else
              c := 128;
          if FBitmapInfo.bmiHeader.biBitCount = 24 then
          begin
            tmp := @PChar(tmpbits)[x*3];
            for y := 0 to AHeight - 1 do
            begin
              tmp[0] := Char(c);
              tmp[1] := Char(c);
              tmp[2] := Char(c);
              Inc(tmp, pitch);
            end;
          end else
            for y := 0 to AHeight - 1 do
              PChar(tmpbits)[y*pitch+x] := Char(c);
        end;
      end;

      z := Height / DestHeight;
      b := Trunc(z * (DestTop-1+0.5));
      for y := 0 to AHeight - 1 do
      begin
        a := Trunc(z * (DestTop+y+0.5));
        if b <> a then
        begin
          b := a;
          if z > 0.25 then
          begin
            if b mod (GridSize*2) = 0 then
              c := 0
            else if b mod GridSize = 0 then
              c := 128
            else
              continue;
          end else
            if b mod GridSize = 0 then
              c := 0
            else
              c := 128;
          FillChar(PChar(tmpbits)[(AHeight-y-1)*pitch], pitch, c);
        end;
      end;
    end;

    SetDIBitsToDevice(hDC, Left, Top, AWidth, AHeight, 0, 0, 0, AHeight,
      tmpbits, PBitmapInfo(@bmi)^, DIB_RGB_COLORS);
    GlobalFree(HGLOBAL(tmpbits));
  end;
end;

procedure TNkVImage.Write(const FileName: WideString; Format: TNkvImageFormat;
  AWidth, AHeight: Integer; ScaleFilter: Boolean);
var
  bfh: TBitmapFileHeader;
  bmi: TMaxBitmapInfo;
  bmisize, bitssize: Integer;
  tmpbits: Pointer;
begin
  bitssize := Integer(CalcPitch(AWidth, FBitmapInfo.bmiHeader.biBitCount))
    * AHeight;
  if (AWidth = Width) and (AHeight = Height) then
  begin
    bmi := FBitmapInfo;
    tmpbits := FBits;
  end else
  begin
    GetMem(tmpbits, bitssize);
    Scale(bmi, tmpbits, AWidth, AHeight, 0, 0, AWidth, AHeight, ScaleFilter);
  end;
  try
    if Format = nifJPEG then
    begin
      GdiplusSaveImage(FileName, gsifJPEG, @bmi, tmpbits);
      Exit;
    end;

    if Format = nifPNG then
    begin
      GdiplusSaveImage(FileName, gsifPNG, @bmi, tmpbits);
      Exit;
    end;

    try
      with TFile.Create(FileName) do
        try
          with bmi do
          begin
            bmisize := SizeOf(bmiHeader);
            if bmiHeader.biBitCount <= 8 then
              Inc(bmisize, (1 shl bmiHeader.biBitCount) * SizeOf(TRGBQUAD));
          end;

          FillChar(bfh, SizeOf(bfh), 0);
          bfh.bfType := Byte('B') + Byte('M') shl 8;
          bfh.bfSize := SizeOf(bfh) + bmisize + bitssize;
          bfh.bfOffBits := SizeOf(bfh) + bmisize;
          bmi.bmiHeader.biSizeImage := bitssize;
          Write(bfh, SizeOf(bfh));
          Write(bmi, bmisize);
          Write(tmpbits^, bitssize);
        finally
          Free;
        end;
    except
      on E: EOSError do raise;
    end;
  finally
    if tmpbits <> FBits then
      FreeMem(tmpbits);
  end;
end;

procedure TNkVImage.GetPixel(X, Y: Integer; var R, G, B: Integer);
var
  p: PChar;
  bpp: Integer;
begin
  bpp := FBitmapInfo.bmiHeader.biBitCount;
  p := @FBits[(Height-Y-1)*Longint(CalcPitch(Width, bpp)) + X*(bpp div 8)];
  if bpp = 24 then
  begin
    B := Ord(p[0]);
    G := Ord(p[1]);
    R := Ord(p[2]);
  end
  else begin
    B := Ord(p[0]);
    G := B;
    R := B;
  end;
end;

end.
